/*
 * Copyright (C) 2001,2002,2003,2005 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

int sock_create(sa_family_t family, int socktype, int proto);
void dump_buffer(char *buff, size_t bufflen, char *prefix);

int main(int argc, char **argv){
	const char *myname = argv[0];
	struct msghdr msghdr;
	struct iovec iov[2];
	char buf[65536], cbuf[2*65536];
	struct sockaddr_storage sa;
	int salen;
	int s = -1, cc;
	int c;
	sa_family_t hints_family = AF_UNSPEC;
	int optval;
	struct addrinfo hints, *res0;
	int gai;
	char *host, *serv;
	char abuf[NI_MAXHOST];
	int opt_help = 0;
	int opt_loop = 0;
	int opt_compat = 0;
	int opt_name;

	printf("Start parsing arguments\n");
	while((c = getopt(argc, argv, "6:"
#if (defined(IPV6_RECVDSTOPTS) || \
     defined(IPV6_RECVHOPOPTS) || \
     defined(IPV6_RECVHOPLIMIT) || \
     defined(IPV6_RECVPKTINFO) || \
     defined(IPV6_RECVRTHDR)) && ( \
     defined(IPV6_2292DSTOPTS) || \
     defined(IPV6_2292HOPOPTS) || \
     defined(IPV6_2292HOPLIMIT) || \
     defined(IPV6_2292PKTINFO) || \
     defined(IPV6_2292RTHDR))
					"c"
#endif
#if defined(IPV6_RECVDSTOPTS) || defined(IPV6_DSTOPTS)
					"d"
#endif
#if defined(IPV6_RECVHOPOPTS) || defined(IPV6_HOPOPTS)
					"h"
#endif
#if defined(IPV6_RECVHOPLIMIT) || defined(IPV6_HOPLIMIT)
					"H"
#endif
#if defined(IPV6_RECVPKTINFO) || defined(IPV6_PKTINFO)
					"i"
#endif
					"l"
#if defined(IPV6_RECVRTHDR) || defined(IPV6_RTHDR)
					"r"
#endif
#if defined(IPV6_RECVTCLASS)
					"t"
#endif
					)) != EOF) {
		switch(c) {
		case '6':
			if (s < 0) {
				s = sock_create(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
				hints_family = AF_INET6;
			}
			if (optarg && *optarg) {
				optval = atoi(optarg);
				printf("set IPV6_V6ONLY to %d\n", optval);
#ifdef IPV6_V6ONLY
				if (setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY, &optval, sizeof(optval)) < 0)
					perror("setsockopt(IPV6_V6ONLY)");
#else
				fprintf(stderr, "IPV6_V6ONLY not supported.\n");
#endif
			}
			if (opt_compat)
				opt_help = 1;
			break;

#if (defined(IPV6_RECVDSTOPTS) || \
     defined(IPV6_RECVHOPOPTS) || \
     defined(IPV6_RECVHOPLIMIT) || \
     defined(IPV6_RECVPKTINFO) || \
     defined(IPV6_RECVRTHDR)) && ( \
     defined(IPV6_2292DSTOPTS) || \
     defined(IPV6_2292HOPOPTS) || \
     defined(IPV6_2292HOPLIMIT) || \
     defined(IPV6_2292PKTINFO) || \
     defined(IPV6_2292RTHDR))
		case 'c':
			opt_compat = 1;
			break;
#endif
#if defined(IPV6_RECVDSTOPTS) || defined(IPV6_DSTOPTS)
		case 'd': /* 2292DSTOPTS / RECVDSTOPTS */
#endif
#if defined(IPV6_RECVHOPOPTS) || defined(IPV6_HOPOPTS)
		case 'h': /* 2292HOPOPTS / RECVHOPOPTS */
#endif
#if defined(IPV6_RECVHOPLIMIT) || defined(IPV6_HOPLIMIT)
		case 'H': /* 2292HOPTLIMIT / RECVHOPLIMIT */
#endif
#if defined(IPV6_RECVPKTINFO) || defined(IPV6_PKTINFO)
		case 'i': /* 2292PKTINFO / RECVPKTINFO */
#endif
#if defined(IPV6_RECVTCLASS)
		case 't': /* RECVTCLASS */
#endif
#if defined(IPV6_RECVRTHDR) || defined(IPV6_RTHDR)
		case 'r': /* 2292RTHDR / RECVRTHDR */
#endif
		    {
			int err;
			int on = 1;

			if (s < 0) {
				s = sock_create(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
				hints_family = AF_INET6;
			}

			switch(c) {
#if defined(IPV6_RECVDSTOPTS) || defined(IPV6_DSTOPTS)
			case 'd':
# if defined(IPV6_2292DSTOPTS) && defined(IPV6_DSTOPTS)
				opt_name = opt_compat ? IPV6_2292DSTOPTS : IPV6_RECVDSTOPTS;
# else
#  if defined(IPV6_RECVDSTOPTS)
				opt_name = IPV6_RECVDSTOPTS;
#  else
				opt_name = IPV6_DSTOPTS;
#  endif
				if (opt_compat)
					opt_help = 1;
# endif
				break;
#endif

#if defined(IPV6_RECVHOPOPTS) || defined(IPV6_HOPOPTS)
			case 'h':
# if defined(IPV6_2292HOPOPTS) && defined(IPV6_HOPOPTS)
				opt_name = opt_compat ? IPV6_2292HOPOPTS : IPV6_RECVHOPOPTS;
# else
#  if defined(IPV6_RECVHOPOPTS)
				opt_name = IPV6_RECVHOPOPTS;
#  else
				opt_name = IPV6_HOPOPTS;
#  endif
				if (opt_compat)
					opt_help = 1;
# endif
				break;
#endif

#if defined(IPV6_RECVHOPLIMIT) || defined(IPV6_HOPLIMIT)
			case 'H':
# if defined(IPV6_2292HOPLIMIT) && defined(IPV6_HOPLIMIT)
				opt_name = opt_compat ? IPV6_2292HOPLIMIT : IPV6_RECVHOPLIMIT;
# else
#  if defined(IPV6_RECVHOPLIMIT)
				opt_name = IPV6_RECVHOPLIMIT;
#  else
				opt_name = IPV6_HOPLIMIT;
#  endif
				if (opt_compat)
					opt_help = 1;
# endif
#endif
				break;

#if defined(IPV6_RECVPKTINFO) || defined(IPV6_PKTINFO)
			case 'i':
# if defined(IPV6_2292PKTINFO) && defined(IPV6_PKTINFO)
				opt_name = opt_compat ? IPV6_2292PKTINFO : IPV6_RECVPKTINFO;
# else
#  if defined(IPV6_RECVPKTINFO)
				opt_name = IPV6_RECVPKTINFO;
#  else
				opt_name = IPV6_PKTINFO;
#  endif
				if (opt_compat)
					opt_help = 1;
# endif
				break;
#endif

#if defined(IPV6_RECVRTHDR) || defined(IPV6_RTHDR)
			case 'r':
# if defined(IPV6_2292RTHDR) && defined(IPV6_RTHDR)
				opt_name = opt_compat ? IPV6_2292RTHDR : IPV6_RECVRTHDR;
# else
#  if defined(IPV6_RECVRTHDR)
				opt_name = IPV6_RECVRTHDR;
#  else
				opt_name = IPV6_RTHDR;
#  endif
				if (opt_compat)
					opt_help = 1;
# endif
				break;
#endif

#if defined(IPV6_RECVTCLASS)
			case 't':
				opt_name = IPV6_RECVTCLASS;
				if (opt_compat)
					opt_help = 1;
				break;
#endif
			}
			opt_compat = 0;
			if (opt_help)
				break;
			err = setsockopt(s, IPPROTO_IPV6, opt_name, &on, sizeof(on));
			if (err)
				warn("failed to setsockopt(): -%c", c);
			break;
			/* 2292HOPOPTS / RECVHOPOPTS */
			opt_compat = 0;
			break;
		    }
		case 'l':
			opt_loop = 1;
			break;
		default:
			opt_help = 1;
		}
	}

	printf("Finished parsing arguments\n");

	argc -= optind;
	argv += optind;

	if (argc == 0 || opt_help) {
		printf("%s [-l] [-6 v6only-sockopt]"
#if defined(IPV6_RECVDSTOPTS) || defined(IPV6_DSTOPTS)
					" ["
#if defined(IPV6_RECVDSTOPTS) && defined(IPV6_2292DSTOPTS)
					" [-c]"
#endif
					"-d]"
#endif
#if defined(IPV6_RECVHOPOPTS) || defined(IPV6_HOPOPTS)
					" ["
#if defined(IPV6_RECVHOPOPTS) && defined(IPV6_2292HOPOPTS)
					" [-c]"
#endif
					"-h]"
#endif
#if defined(IPV6_RECVHOPLIMIT) || defined(IPV6_HOPLIMIT)
					" ["
#if defined(IPV6_RECVHOPLIMIT) && defined(IPV6_2292HOPLIMIT)
					" [-c]"
#endif
					"-H]"
#endif
#if defined(IPV6_RECVPKTINFO) || defined(IPV6_PKTINFO)
					" ["
#if defined(IPV6_RECVPKTINFO) && defined(IPV6_2292PKTINFO)
					" [-c]"
#endif
					"-i]"
#endif
#if defined(IPV6_RECVRTHDR) || defined(IPV6_RTHDR)
					" ["
#if defined(IPV6_RECVRTHDR) && defined(IPV6_2292RTHDR)
					" [-c]"
#endif
					"-r]"
#endif
#if defined(IPV6_RECVTCLASS)
					" ["
					"-t]"
#endif
			" [host] port\n"
			"\n"
#if defined(IPV6_RECVDSTOPTS) || defined(IPV6_DSTOPTS)
# if defined(IPV6_RECVDSTOPTS)
			"\t-d   : enable IPV6_RECVDSTOPTS [3542]\n"
#  if defined(IPV6_2292DSTOPTS)
			"\t-c -d: enable IPV6_2292DSTOPTS (IPV6_DSTOPTS [2292])\n"
#  endif
# else
			"\t-d   : enable IPV6_DSTOPTS [2292]\n"
# endif
#endif
#if defined(IPV6_RECVHOPOPTS) || defined(IPV6_HOPOPTS)
# if defined(IPV6_RECVHOPOPTS) || defined(IPV6_2292HOPOPTS)
			"\t-c -h: enable IPV6_2292HOPOPTS (IPV6_HOPOPTS [2292])\n"
			"\t-h   : enable IPV6_RECVHOPOPTS [3542]\n"
# else
#  if defined(IPV6_RECVDSTOPTS)
			"\t-h   : enable IPV6_RECVHOPOPTS [3542]\n"
#  else
			"\t-h   : enable IPV6_HOPOPTS [2292]\n"
#  endif
# endif
#endif
#if defined(IPV6_RECVHOPLIMIT) || defined(IPV6_HOPLIMIT)
# if defined(IPV6_RECVHOPLIMIT) || defined(IPV6_2292HOPLIMIT)
			"\t-c -H: enable IPV6_2292HOPLIMIT (IPV6_HOPLIMIT [2292])\n"
			"\t-H   : enable IPV6_RECVHOPLIMIT [3542]\n"
# else
#  if defined(IPV6_RECVDSTOPTS)
			"\t-H   : enable IPV6_RECVHOPLIMIT [3542]\n"
#  else
			"\t-H   : enable IPV6_HOPLIMIT [2292]\n"
#  endif
# endif
#endif
#if defined(IPV6_RECVPKTINFO) || defined(IPV6_PKTINFO)
# if defined(IPV6_RECVPKTINFO) || defined(IPV6_2292PKTINFO)
			"\t-c -i: enable IPV6_2292PKTINFO (IPV6_PKTINFO [2292])\n"
			"\t-i   : enable IPV6_RECVPKTINFO [3542]\n"
# else
#  if defined(IPV6_RECVDSTOPTS)
			"\t-i   : enable IPV6_RECVPKTINFO [3542]\n"
#  else
			"\t-i   : enable IPV6_PKTINFO [2292]\n"
#  endif
# endif
#endif
#if defined(IPV6_RECVRTHDR) || defined(IPV6_RTHDR)
# if defined(IPV6_RECVRTHDR) || defined(IPV6_2292RTHDR)
			"\t-c -r: enable IPV6_2292RTHDR (IPV6_RTHDR [2292])\n"
			"\t-r   : enable IPV6_RECVRTHDR [3542]\n"
# else
#  if defined(IPV6_RECVDSTOPTS)
			"\t-r   : enable IPV6_RECVRTHDR [3542]\n"
#  else
			"\t-r   : enable IPV6_RTHDR [2292]\n"
#  endif
# endif
#endif
#if defined(IPV6_RECVTCLASS)
			"\t-t   : enable IPV6_RECVTCLASS [3542]\n"
#endif
			"",
			myname);
		exit(1);
	}

        host = argc > 1 ? argv[0] : NULL;
        serv = argc > 1 ? argv[1] : argv[0];

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = hints_family;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_flags = AI_PASSIVE;
	gai = getaddrinfo(host, serv, &hints, &res0);
	if (gai) {
		printf("getaddrinfo(): %s\n",
			gai_strerror(gai));
		exit(2);
	}

	if (s < 0) {
		s = sock_create(res0->ai_family, res0->ai_socktype, res0->ai_protocol);
	} else if (hints_family != res0->ai_family) {
		errx(1, "getaddrinfo() returns inappropriate family %d (!= %d)",
		     (int)res0->ai_family, (int)hints_family);
	}

	if (bind(s, res0->ai_addr, res0->ai_addrlen)) {
		freeaddrinfo(res0);
		perror("bind");
		exit(2);
	}
	freeaddrinfo(res0);

	do {
		struct cmsghdr *cmsg;

		memset(&iov, 0, sizeof(iov));
		iov[0].iov_base = buf;
		iov[0].iov_len = sizeof(buf);

		memset(&msghdr, 0, sizeof(msghdr));
		msghdr.msg_name = (void *)&sa;
		msghdr.msg_namelen = sizeof(sa);
		msghdr.msg_iov = iov;
		msghdr.msg_iovlen = 1;
		msghdr.msg_control = cbuf;
		msghdr.msg_controllen = sizeof(cbuf);
		msghdr.msg_flags = 0;

		cc = recvmsg(s, &msghdr, 0);
		if (cc < 0) {
			perror("recvmsg");
			exit(2);
		}
		salen = msghdr.msg_namelen;

		if (getnameinfo((struct sockaddr *)&sa, salen,
				abuf, sizeof(abuf), NULL, 0,
				NI_NUMERICHOST))
			strcpy(abuf, "???");

		printf("received %d byte(s) from %s\n", cc, abuf);

		for (cmsg = CMSG_FIRSTHDR(&msghdr); cmsg;
		     cmsg = CMSG_NXTHDR(&msghdr, cmsg)) {
			printf("\tcmsg(level=%d, type=%d)\n", cmsg->cmsg_level, cmsg->cmsg_type);
			dump_buffer(CMSG_DATA(cmsg), cmsg->cmsg_len, "\t\t");
		}
	} while(opt_loop);

	exit(0);
}

int sock_create(sa_family_t family, int socktype, int proto)
{
	int s;

	printf("Creating socket (family = %d) ...", family);
	s = socket(family, socktype, proto);
	if (s < 0) {
		perror("socket");
		exit(2);
	}
	printf(" %d\n", s);
	return s;
}

void dump_buffer(char *buff, size_t bufflen, char *prefix)
{
	size_t offset;

	for (offset = 0; offset < bufflen; offset++) {
		if ((offset & 15) == 0)
			printf("%s%08x:", prefix, (unsigned int)offset);
		printf(" %02x", buff[offset] & 0xff);
		if ((offset & 15) == 15 || offset == bufflen - 1)
			printf("\n");
	}
}

